/* palace-control.c - global controluration for Palace DCI
 *
 * $Id: palace-control.c, v1.0 2002/06/07 18:25:16 nitroglycerine Exp $
 * 
 *     Low-level device control interface.  Will become mid-level device
 * control interface to talk between Palace and low-level hardware plugins.
 */

#include "palace.h"

/* Headers specific to ppdev output */
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/ppdev.h>


gboolean send_ok = FALSE;	/* Have perms to write to port? */

/* Switch to a struct when turning into a generic plugin */
static int ppdfd = 0;
static int devicenum = 0;
static int ppraw_enabled = 0;

/* TODO: Make this a separate configuration option */
static char devicename[3][16] = {
	"/dev/parport0",	/* 0x378 */
	"/dev/parport1",	/* 0x278 */
	"/dev/parport2"		/* 0x3BC */
};

static inline void port_out(unsigned short port, unsigned char value)
{
	__asm__ volatile (
		"outb %0, %1"
		:
		: "a" ((char)value),
		  "d"((unsigned short)port)
	);
}


static inline unsigned char port_in(unsigned short port)
{
	char _v;
	__asm__ volatile (
	 	"inb %1, %0"
		: "=a" ((char)_v)
		: "d" ((unsigned short)port)
	);
	return _v;
}

/* Initialize parallel port output using ppdev - nonzero on error */
int palace_control_init_ppdev() {
	unsigned char c = 0;
	char *devname = devicename[devicenum];

	switch(palacecfg.lpt_address) {
		case 0x378:
			devicenum = 0;
			break;
		case 0x278:
			devicenum = 1;
			break;
		case 0x3BC:
			devicenum = 2;
			break;
		default:
			devicenum = 0;
			break;
	}

	if(ppdfd) {
		/* Already initialized */
		debug_printf("Palace error: ppdev output already initialized\n");
		return -1;
	}

	/* Open the port */
	if((ppdfd = open(devname, O_RDWR)) < 0) {
		fprintf(stderr, "Palace: Unable to open %s: %s\n", devname, strerror(errno));
		ppdfd = 0;
		return -1;
	}

	/* Request exclusive access to the ppdev device */
	if(ioctl(ppdfd, PPEXCL) < 0) {
		debug_printf("Palace: Unable to request exclusive access to parallel port: %s\n", strerror(errno));
		palace_control_cleanup_ppdev();
		return -1;
	}

	/* Claim driver access to the ppdev device */
	if(ioctl(ppdfd, PPCLAIM) < 0) {
		debug_printf("Palace: Unable to claim the parallel port: %s\n", strerror(errno));
		palace_control_cleanup_ppdev();
		return -1;
	}

	/* Set the data pins to zero */
	if(ioctl(ppdfd, PPWDATA, &c) < 0) {
		debug_printf("Palace: Unable to set parallel port to zero: %s\n", strerror(errno));
	}

	return 0;
}

/* Clean up ppdev output */
void palace_control_cleanup_ppdev() {
	unsigned char c = 0;
	
	if(!ppdfd) {
		return;
	}

	/* Set the data pins to zero */
	if(ioctl(ppdfd, PPWDATA, &c) < 0) {
		debug_printf("Palace: Unable to set parallel port back to zero: %s\n", strerror(errno));
	}

	/* Release driver access to the ppdev device */
	if(ioctl(ppdfd, PPRELEASE, &c) < 0) {
		debug_printf("Palace: Unable to release the parallel port: %s\n", strerror(errno));
	}

	if(close(ppdfd)) {
		debug_printf("Palace: Unable to close %s: %s\n", devicename[devicenum], strerror(errno));
	}

	ppdfd = 0;
	devicenum = 0;
}

/* Output using ppdev */
void palace_control_output_ppdev(unsigned char value)
{
	if(!ppdfd) {
		return;
	}

	if(ioctl(ppdfd, PPWDATA, &value) < 0) {
		debug_printf("Unable to output to ppdev: %s\n", strerror(errno));
	}

	/* I don't know how fast the parallel port can respond */
	usleep(1000);
}

/* Initialize raw parallel port output - nonzero on error */
int palace_control_init_ppraw() {
	if(ioperm(palacecfg.lpt_address, 3, 1)) {
		perror("Palace: Unable to get LPT port I/O permission (root privileges required)");
		ppraw_enabled = 0;
		return -1;
	}

	ppraw_enabled = 1;

	return 0;
}

/* Clean up raw parallel port output */
void palace_control_cleanup_ppraw() {
	if(!ppraw_enabled) {
		return;
	}
	palace_control_output_ppraw(0);
	if (ioperm(palacecfg.lpt_address, 3, 0)) {
		debug_printf("Palace: Error releasing LPT port I/O permission: %s\n", strerror(errno));
	}
	ppraw_enabled = 0;
}

/* Output using ppraw */
void palace_control_output_ppraw(unsigned char value) {
	if(!ppraw_enabled) {
		debug_printf("Palace: Attempt to output to ppraw when not enabled\n");
		return;
	}
	port_out(palacecfg.lpt_address, value);
}

int palace_control_init() {
	/*
	if(!palace_control_init_ppdev()) {
		return 0;
	} else */if(!palace_control_init_ppraw()) {
		return 0;
	}

	return -1;
}

void palace_control_cleanup() {
	if(ppdfd) { 
		palace_control_cleanup_ppdev();
	} else {
		palace_control_cleanup_ppraw();
	}
}

void palace_control_output(unsigned char value) {
	if(ppdfd) {
		palace_control_output_ppdev(value);
	} else {
		palace_control_output_ppraw(value);
	}
}

